#define	HOST_ID	"www.cwo.com"
//#define	HOST_ID	"www.yahoo.com"
//#define	HOST_ID	"www.symmetricom.com"
//#define	HOST_ID	"nguyen"
//#define	HOST_ID	"192.168.47.31"

#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif

#include <winsock.h>
#include <stdio.h>
#include <stdlib.h>

///////////////////////////////////////////////////////////////////////////////
//#include "iphdr.h"

// Align on a 1-byte boundary
#include <pshpack1.h>

#ifndef UINT8
#define	UINT8 unsigned char
#endif

#ifndef UINT16
#define	UINT16 unsigned short
#endif

#ifndef UINT32
#define	UINT32 unsigned	// int, long
#endif

// IPv4 header
typedef struct _IP_HDR
{
    UINT8  ip_verlen;        // 4-bit IPv4 version
                                     // 4-bit header length (in 32-bit words)
    UINT8  ip_tos;           // IP type of service
    UINT16 ip_totallength;   // Total length
    UINT16 ip_id;            // Unique identifier
    UINT16 ip_offset;        // Fragment offset field
    UINT8  ip_ttl;           // Time to live
    UINT8  ip_protocol;      // Protocol(TCP,UDP etc)
    UINT16 ip_checksum;      // IP checksum
    UINT32   ip_srcaddr;       // Source address
    UINT32   ip_destaddr;      // Source address
} IP_HDR;

// IPv4 option header
typedef struct _IP_OPTION_HDR
{
    UINT8   opt_code;           // option type
    UINT8   opt_len;            // length of the option header
    UINT8   opt_ptr;            // offset into options
    UINT32   opt_addr[9];        // list of IPv4 addresses
} IP_OPTION_HDR;

// ICMP header
typedef struct _ICMP_HDR
{
    UINT8   icmp_type;
    UINT8   icmp_code;
    UINT16  icmp_checksum;
    UINT16  icmp_id;
    UINT16  icmp_sequence;
    UINT32   icmp_timestamp;
} ICMP_HDR;

// IPv4 option for record route
#define IP_RECORD_ROUTE     0x7

// ICMP types and codes
#define ICMP_ECHO_REQUEST_TYPE   8
#define ICMP_ECHO_REQUEST_CODE   0
#define ICMP_ECHO_REPLY_TYPE     0
#define ICMP_ECHO_REPLY_CODE     0
#define ICMP_MINIMUM_HEADER      8

// Restore byte alignment back to default
#include <poppack.h>

///////////////////////////////////////////////////////////////////////////////

#define DEFAULT_DATA_SIZE      32       // default data size

#define DEFAULT_SEND_COUNT     4        // number of ICMP requests to send (PING)

#define DEFAULT_RECV_TIMEOUT   6000     // six second

#define DEFAULT_TTL            30		// PING (128) TRACERT (30)

//BOOL  bResolve       = FALSE;         	// Resolve addresses to host names (TRACE ONLY)
BOOL  bResolve       = TRUE;         	// Resolve addresses to host names (TRACE ONLY)

BOOL  bRecordRoute = TRUE;          	// Use IPv4 record route?
//BOOL  bRecordRoute = FALSE;       	// Use IPv4 record route?

//
// Function: InitIcmpHeader
//
// Description:
//    Helper function to fill in various stuff in our ICMP request.
//
void InitIcmpHeader(char *buf, int datasize)
{
    ICMP_HDR   *icmp_hdr=NULL;
    char       *datapart=NULL;

    icmp_hdr = (ICMP_HDR *)buf;
    icmp_hdr->icmp_type     = ICMP_ECHO_REQUEST_TYPE;        // request an ICMP echo
    icmp_hdr->icmp_code     = ICMP_ECHO_REQUEST_CODE;
    icmp_hdr->icmp_id       = (UINT16)GetCurrentProcessId();
    icmp_hdr->icmp_checksum = 0;
    icmp_hdr->icmp_sequence = 0;
    icmp_hdr->icmp_timestamp= GetTickCount();

    datapart = buf + sizeof(ICMP_HDR);
    //
    // Place some junk in the buffer.
    //
    memset(datapart, 'E', datasize);
}

//
// Function: checksum
//
// Description:
//    This function calculates the 16-bit one's complement sum
//    of the supplied buffer (ICMP) header.
//
UINT16 checksum(UINT16 *buffer, int size)
{
    UINT32 cksum=0;

    while (size > 1)
    {
        cksum += *buffer++;
        size -= sizeof(UINT16);
    }
    if (size)
    {
        cksum += *(UCHAR*)buffer;
    }
    cksum = (cksum >> 16) + (cksum & 0xffff);
    cksum += (cksum >>16);
    return (UINT16)(~cksum);
}

//
// Function: ComputeIcmpChecksum
//
// Description:
//    This routine computes the checksum for the ICMP request for the ICMP packet and data
//
void ComputeIcmpChecksum(SOCKET s, char *buf, int packetlen)
{
    ULONG    sequence=0;

    sequence = GetTickCount();

	ICMP_HDR    *icmpv4=NULL;

	icmpv4 = (ICMP_HDR *)buf;

	icmpv4->icmp_sequence = (UINT16)sequence;

	icmpv4->icmp_checksum = 0;
	icmpv4->icmp_checksum = checksum((UINT16 *)buf, packetlen);
}

//
// Function: SetTtl
//
// Description:
//    Sets the TTL on the socket.
//
int SetTtl(SOCKET s, int ttl)
{
    int     optlevel,
            option,
            rc;

    rc = NO_ERROR;

	optlevel = IPPROTO_IP;
	option   = IP_TTL;

    if (rc == NO_ERROR)
    {
        rc = setsockopt(
                s,
                optlevel,
                option,
                (char *)&ttl,
                sizeof(ttl)
                );
    }

    if (rc == SOCKET_ERROR)
    {
        fprintf(stderr, "SetTtl: setsockopt failed: %d\n", WSAGetLastError());
    }
    return rc;
}

//
// Function: PrintPayload
//
// Description:
//    It determines if there are any IP options
//    present (by seeing if the IP header length is greater than 20 bytes) and
//    if so it prints the IP record route options.
//
void PrintPayload(char *buf)
{
    int     hdrlen=0,
            routes=0,
            i;

	LPHOSTENT 		lpHost=NULL;

	SOCKADDR_IN      hop;
	IP_OPTION_HDR *v4opt=NULL;
	IP_HDR        *v4hdr=NULL;

	hop.sin_family = (UINT16)AF_INET;
	hop.sin_port   = 0;

	v4hdr = (IP_HDR *)buf;
	hdrlen = (v4hdr->ip_verlen & 0x0F) * 4;

	// If the header length is greater than the size of the basic IPv4
	//    header then there are options present. Find them and print them.
	if (hdrlen > sizeof(IP_HDR))
	{
		v4opt = (IP_OPTION_HDR *)(buf + sizeof(IP_HDR));
		routes = (v4opt->opt_ptr / sizeof(ULONG)) - 1;
		for(i=0; i < routes ;i++)
		{
			hop.sin_addr.s_addr = v4opt->opt_addr[i];

			// Print the route
			if (i == 0)
				printf("    Route: ");
			else
				printf("           %d: ", i);
			printf("%s ", inet_ntoa(hop.sin_addr));
			if (bResolve)
				lpHost = gethostbyaddr((char*)&hop.sin_addr, 4, AF_INET);

			if (lpHost && (i < routes-1))
				printf(" [%s] ->\n", lpHost->h_name);
			else
				printf("\n");
		}
	}

    return;
}

//
// Function: main
//
// Description:
//    Setup the ICMP raw socket and create the ICMP header. Add
//    the appropriate IP option header and start sending ICMP
//    echo requests to the endpoint.
//
int main(void)
{

    WSADATA            wsd;
    SOCKET             s=INVALID_SOCKET;
    char              *icmpbuf=NULL,
                       recvbuf[0xFFFF];
    int                packetlen=0,
                       recvbuflen=0xFFFF,
                       time=0,
                       rc,
                       i;
    IP_OPTION_HDR    ipopt;

	LPHOSTENT lpHost;
	SOCKADDR_IN saDest;
    SOCKADDR_IN   	saSrc;
	int 			nAddrLen = sizeof(struct sockaddr_in);

    ///////////////////////////////////////////////////////////////////////////
    // Load Winsock lib
    if (WSAStartup(MAKEWORD(1,1), &wsd) != 0)
    {
        printf("WSAStartup() failed: %d\n", GetLastError());
        return -1;
    }

    ///////////////////////////////////////////////////////////////////////////
    // Create the raw socket
    s = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
    if (s == INVALID_SOCKET)
    {
        printf("socket failed: %d\n", WSAGetLastError());
        return -1;
    }

    ///////////////////////////////////////////////////////////////////////////
	// Lookup host
	lpHost = gethostbyname(HOST_ID);
	if (lpHost == NULL)
	{
		fprintf(stderr,"\nHost not found: %s\n", HOST_ID);
		return -1;
	}

	// Setup destination socket address
	saDest.sin_addr.s_addr = *((u_long FAR *) (lpHost->h_addr));
	saDest.sin_family = AF_INET;
	saDest.sin_port = 0;

    ///////////////////////////////////////////////////////////////////////////
    // Figure out the size of the ICMP header and payload
	packetlen += sizeof(ICMP_HDR);

    // Add in the data size
    packetlen += DEFAULT_DATA_SIZE;

    // Allocate the buffer that will contain the ICMP request
    icmpbuf = (char *)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, packetlen);
    if (icmpbuf == NULL)
    {
        fprintf(stderr, "HeapAlloc failed: %d\n", GetLastError());
        return -1;
    }

    // Initialize the ICMP headers
	if (bRecordRoute)
	{
		// Setup the IP option header to go out on every ICMP packet
		ZeroMemory(&ipopt, sizeof(ipopt));
		ipopt.opt_code = IP_RECORD_ROUTE; // record route option
		ipopt.opt_ptr  = 4;               // point to the first addr offset
		ipopt.opt_len  = 39;              // length of option header

		rc = setsockopt(s, IPPROTO_IP, IP_OPTIONS,
				(char *)&ipopt, sizeof(ipopt));
		if (rc == SOCKET_ERROR)
		{
			fprintf(stderr, "setsockopt(IP_OPTIONS) failed: %d\n", WSAGetLastError());
			return -1;
		}
	}
	InitIcmpHeader(icmpbuf, DEFAULT_DATA_SIZE);

    printf("\nPinging ");
    printf("%s [%s] ", inet_ntoa(saDest.sin_addr), HOST_ID);
    printf(" with %d bytes of data\n\n", DEFAULT_DATA_SIZE);

	SetTtl(s, DEFAULT_TTL);

    ///////////////////////////////////////////////////////////////////////////
   // Start sending the ICMP requests
    for(i=0; i < DEFAULT_SEND_COUNT ;i++)
    {
        // Set the sequence number and compute the checksum
        ComputeIcmpChecksum(s, icmpbuf, packetlen);

        time = GetTickCount();
        rc = sendto(
                s,
                icmpbuf,
                packetlen,
                0,
                (LPSOCKADDR)&saDest /* destination */,
                sizeof(SOCKADDR_IN) /* address length */
                );
        if (rc == SOCKET_ERROR)
        {
            fprintf(stderr, "sendto failed: %d\n", WSAGetLastError());
            return -1;
        }


		if (i < DEFAULT_SEND_COUNT)
		{
			if(recvfrom(s, recvbuf, recvbuflen, 0, (LPSOCKADDR)&saSrc, &nAddrLen) < 0) {
				fprintf(stderr,"<< ERR >> recvfrom() failed: %d\n", GetLastError());
				return -1;
			}
		}

		time = time - GetTickCount();


		printf("Reply from : ");

			printf("%s ", inet_ntoa(saSrc.sin_addr));
		if (time == 0)
			printf(": bytes=%d time<1ms TTL=%d\n", DEFAULT_DATA_SIZE, DEFAULT_TTL);
		else
			printf(": bytes=%d time=%dms TTL=%d\n", DEFAULT_DATA_SIZE, time, DEFAULT_TTL);

		PrintPayload(recvbuf);

    }

    //
    // Cleanup
    //
    if (s != INVALID_SOCKET)
        closesocket(s);
    HeapFree(GetProcessHeap(), 0, icmpbuf);

    WSACleanup();
    return 0;
}

