#define	HOST_ID	"www.cwo.com"
//#define	HOST_ID	"www.yahoo.com"
//#define	HOST_ID	"192.168.47.21"

#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif

#include <winsock2.h>
#include <ws2tcpip.h>
#include <stdio.h>
#include <stdlib.h>

//#include "resolve.h"

///////////////////////////////////////////////////////////////////////////////
//#include "iphdr.h"

// Set alignment boundary to 1 byte
#include <pshpack1.h>

#ifndef UINT8
#define	UINT8 unsigned char
#endif

#ifndef UINT16
#define	UINT16 unsigned short
#endif

#ifndef UINT32
#define	UINT32 unsigned	// int, long
#endif


// IPv4 header
typedef struct _IP_HDR
{
    UINT8  ip_verlen;        // 4-bit IPv4 version
                                     // 4-bit header length (in 32-bit words)
    UINT8  ip_tos;           // IP type of service
    UINT16 ip_totallength;   // Total length
    UINT16 ip_id;            // Unique identifier
    UINT16 ip_offset;        // Fragment offset field
    UINT8  ip_ttl;           // Time to live
    UINT8  ip_protocol;      // Protocol(TCP,UDP etc)
    UINT16 ip_checksum;      // IP checksum
    UINT32   ip_srcaddr;       // Source address
    UINT32   ip_destaddr;      // Source address
} IP_HDR;

// IPv4 option header
typedef struct _IP_OPTION_HDR
{
    UINT8   opt_code;           // option type
    UINT8   opt_len;            // length of the option header
    UINT8   opt_ptr;            // offset into options
    UINT32   opt_addr[9];        // list of IPv4 addresses
} IP_OPTION_HDR;

// ICMP header
typedef struct _ICMP_HDR
{
    UINT8   icmp_type;
    UINT8   icmp_code;
    UINT16  icmp_checksum;
    UINT16  icmp_id;
    UINT16  icmp_sequence;
    UINT32   icmp_timestamp;
} ICMP_HDR;

#define IP_RECORD_ROUTE     0x7

// ICMP types and codes
#define ICMP_ECHO_REQUEST_TYPE   8
#define ICMP_ECHO_REQUEST_CODE   0
#define ICMP_ECHO_REPLY_TYPE     0
#define ICMP_ECHO_REPLY_CODE     0

#define ICMP_DESTUNREACH    3
#define ICMP_SRCQUENCH      4
#define ICMP_REDIRECT       5
#define ICMP_ECHO           8
#define ICMP_TIMEOUT       11
#define ICMP_PARMERR       12

// Restore byte alignment to compile default
#include <poppack.h>

///////////////////////////////////////////////////////////////////////////////

#define DEFAULT_DATA_SIZE      32       // default data size

#define DEFAULT_SEND_COUNT     4        // number of ICMP requests to send (PING)

#define DEFAULT_RECV_TIMEOUT   6000     // six second

#define DEFAULT_TTL            60		// PING (128) TRACERT (30)

BOOL  bResolve       = TRUE;              // Resolve addresses to host names (TRACE ONLY)

//
// Function: InitIcmpHeader
//
// Description:
//    Helper function to fill in various stuff in our ICMP request.
//
void InitIcmpHeader(char *buf, int datasize)
{
    ICMP_HDR   *icmp_hdr=NULL;
    char       *datapart=NULL;

    icmp_hdr = (ICMP_HDR *)buf;
    icmp_hdr->icmp_type     = ICMP_ECHO_REQUEST_TYPE;        // request an ICMP echo
    icmp_hdr->icmp_code     = ICMP_ECHO_REQUEST_CODE;
    icmp_hdr->icmp_id       = (UINT16)GetCurrentProcessId();
    icmp_hdr->icmp_checksum = 0;
    icmp_hdr->icmp_sequence = 0;
    icmp_hdr->icmp_timestamp= GetTickCount();

    datapart = buf + sizeof(ICMP_HDR);
    //
    // Place some junk in the buffer.
    //
    memset(datapart, 'E', datasize);
}

///////////////////////////////////////////////////////////////////////////////
/* Do TCP-style checksum. Improved algorithm is from RFC 1071 */
//u_short csum(void *dp, u_short count)
UINT16 csum(UINT16 *dp, int count)
{
    register u_long total=0L;
    register u_short n, *p, carries;

    n = count / 2;
    p = (u_short *)dp;
    while (n--)
        total += *p++;
    if (count & 1)
        total += *(BYTE *)p;
    while ((carries=(u_short)(total>>16))!=0)
        total = (total & 0xffffL) + carries;
    return (u_short)(~total);
}

//
// Function: ComputeIcmpChecksum
//
// Description:
//    This routine computes the checksum for the ICMP request for the ICMP packet and data
//
void ComputeIcmpChecksum(char *buf, int packetlen, UINT16 sequence)
{
    {
        ICMP_HDR    *icmpv4=NULL;

        icmpv4 = (ICMP_HDR *)buf;
        icmpv4->icmp_sequence = (UINT16)sequence;
        icmpv4->icmp_checksum = 0;
//        icmpv4->icmp_checksum = checksum((UINT16 *)buf, packetlen);
        icmpv4->icmp_checksum = csum((UINT16 *)buf, packetlen);
    }
}

//
// Function: SetTtl
//
// Description:
//    Sets the TTL on the socket.
//
int SetTtl(SOCKET s, int ttl)
{
    int     optlevel,
            option,
            rc;

    rc = NO_ERROR;
//    if (AF_INET == AF_INET)
    {
        optlevel = IPPROTO_IP;
        option   = IP_TTL;
    }

//    if (rc == NO_ERROR)
    {
        rc = setsockopt(
                s,
                optlevel,
                option,
                (char *)&ttl,
                sizeof(ttl)
                );
    }

    if (rc == SOCKET_ERROR)
    {
        fprintf(stderr, "SetTtl: setsockopt failed: %d\n", WSAGetLastError());
    }
    return rc;
}

//
// Function: IsSockaddrEqual
//
// Description:
//    This routines compares two SOCKADDR structure to determine
//    whether the address portion of them are equal. Zero is returned
//    for equal; non-zero for not equal.
//
int IsSockaddrEqual(SOCKADDR *sa1, SOCKADDR *sa2)
{
    int rc;

    rc = 1;
    if (sa1->sa_family == sa2->sa_family)
    {
        if (sa1->sa_family == AF_INET)
        {
            rc = memcmp(
                    &((SOCKADDR_IN *)sa1)->sin_addr,
                    &((SOCKADDR_IN *)sa2)->sin_addr,
                    sizeof(struct in_addr)
                    );
            rc = rc;
        }
    }

    return rc;
}

//
// Function: AnalyzePacket
//
// Description:
//    This routines finds the ICMP packet within the encapsulated header and
//    verifies that the ICMP packet is a TTL expired or echo reply message.
//    If not then an error is returned.
//
int AnalyzePacket(char *buf)
{
    int     hdrlen=0,
            routes=0,
            rc;

    rc = NO_ERROR;

    {
        IP_HDR        *v4hdr=NULL;
        ICMP_HDR        *icmphdr=NULL;

        v4hdr = (IP_HDR *)buf;
        hdrlen = (v4hdr->ip_verlen & 0x0F) * 4;

        if (v4hdr->ip_protocol == IPPROTO_ICMP)
        {
            icmphdr = (ICMP_HDR *)&buf[hdrlen];

            if ((icmphdr->icmp_type != ICMP_TIMEOUT) &&
                    (icmphdr->icmp_type != ICMP_ECHO_REPLY_TYPE) &&
                    (icmphdr->icmp_code != ICMP_ECHO_REPLY_CODE) )
            {
                printf("Received ICMP message type %d instead of TTL expired!\n", icmphdr->icmp_type);
                rc = SOCKET_ERROR;
            }
        }
    }

    return rc;
}

//
// Function: main
//
// Description:
//    Setup the ICMP raw socket and create the ICMP header. Add
//    the appropriate IP option header and start sending ICMP
//    echo requests to the endpoint.
//
//int __cdecl main(int argc, char **argv)
int main(void)
{

    WSADATA         wsd;
    SOCKET          s=INVALID_SOCKET;
    char           *icmpbuf=NULL,
                    recvbuf[0xFFFF];
    int             packetlen=0,
                    recvbuflen=0xFFFF,
                    notdone,				// TRACE ONLY
                    ttl;

	unsigned    	time;

	LPHOSTENT 		lpHost;
	SOCKADDR_IN 	saDest;
    SOCKADDR_IN   	saSrc;
	int 			nAddrLen = sizeof(struct sockaddr_in);

	UINT16    sequence;

    ///////////////////////////////////////////////////////////////////////////
    // Load Winsock Lib
    if (WSAStartup(MAKEWORD(2,2), &wsd) != 0)
    {
        printf("WSAStartup() failed: %d\n", GetLastError());
        return -1;
    }

    ///////////////////////////////////////////////////////////////////////////
	// Lookup host
	lpHost = gethostbyname(HOST_ID);
	if (lpHost == NULL)
	{
		fprintf(stderr,"\nHost not found: %s\n", HOST_ID);
		return -1;
	}

	// Setup destination socket address
	saDest.sin_addr.s_addr = *((u_long FAR *) (lpHost->h_addr));	// 0x292fa8c0
	saDest.sin_family = AF_INET;
	saDest.sin_port = 0;

    ///////////////////////////////////////////////////////////////////////////
    // Create the raw socket
    s = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
    if (s == INVALID_SOCKET)
    {
        printf("socket failed: %d\n", WSAGetLastError());
        return -1;
    }

    ///////////////////////////////////////////////////////////////////////////
    // Figure out the size of the ICMP header and payload
	packetlen += sizeof(ICMP_HDR);

    // Add in the data size
    packetlen += DEFAULT_DATA_SIZE;

    // Allocate the buffer that will contain the ICMP request
    icmpbuf = (char *)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, packetlen);
    if (icmpbuf == NULL)
    {
        fprintf(stderr, "HeapAlloc failed: %d\n", GetLastError());
        return -1;
    }

    // Initialize the ICMP headers
	InitIcmpHeader(icmpbuf, DEFAULT_DATA_SIZE);


    printf("\nTraceroute to %s [%s] over a maximum of %d hops\n\n", inet_ntoa(saDest.sin_addr), HOST_ID, DEFAULT_TTL);

    ttl = 1;

	sequence = (UINT16)GetTickCount();

    ///////////////////////////////////////////////////////////////////////////
    // Start sending the ICMP requests
    do
    {
        notdone = 1;

        SetTtl(s, ttl);

        // Set the sequence number and compute the checksum
        ComputeIcmpChecksum(icmpbuf, packetlen, sequence++);

        // Send the ICMP echo request
        time = GetTickCount();

        if (sendto(
                s,
                icmpbuf,
                packetlen,
                0,
                (LPSOCKADDR)&saDest /* destination */,
                sizeof(SOCKADDR_IN) /* address length */
                ) != packetlen)
        {
            fprintf(stderr, "sendto failed: %d\n", WSAGetLastError());
            return -1;
        }


		// Post another receive
		if (notdone)
		{
			recvbuf[0] = '\0';
			if(recvfrom(s, recvbuf, recvbuflen, 0, (LPSOCKADDR)&saSrc, &nAddrLen) < 0) {
				fprintf(stderr,"<< ERR >> recvfrom() failed: %d\n", GetLastError());
				return -1;
			}
		}

		if (AnalyzePacket(recvbuf) == NO_ERROR)
		{
			lpHost = gethostbyaddr((char*)&saSrc.sin_addr, 4, AF_INET);

			time = GetTickCount() - time;

			if(lpHost)
				printf("%d   %d ms   %s [%s]\n", ttl, time, inet_ntoa(saSrc.sin_addr), lpHost->h_name);
			else
				printf("%d   %d ms   %s\n", ttl, time, inet_ntoa(saSrc.sin_addr));

			// See if the response is from the desired destination
			notdone = IsSockaddrEqual((SOCKADDR *)&saDest, (SOCKADDR *)&saSrc);

			// Increment the TTL
			ttl++;
		}

//        Sleep(1000);
    } while ((notdone) && (ttl < DEFAULT_TTL));

    //
    // Cleanup
    //

    if (s != INVALID_SOCKET)
        closesocket(s);

    HeapFree(GetProcessHeap(), 0, icmpbuf);

    WSACleanup();
    return 0;
}


